! *****************************************************************
!
! This file contains a Fortran 90 module for setting up and computing 
! the analytic solution.  
!
! Copyright (C) 2009 David McInerney, University of Chicago
!
! *****************************************************************
!
!    This program is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with this program (see file "COPYING").  If not, see
!    <http://www.gnu.org/licenses/>.
!
! *****************************************************************

module analytic

! *****************************************************************

  use parameters, only : period, g, h0, Cl, LL, BB, tau_0, rho, angle, &
       sigma, alpha
  use grid, only : JJ, KK, x_at_zeta, y_at_zeta, x_at_u, &
       y_at_u, x_at_v, y_at_v

  integer, parameter :: NN = 31
!  integer, parameter :: NN = 30

  double complex, dimension(1:JJ,1:KK) :: ZZ_bar_at_zeta, ZZ_hat_at_zeta, &
       PP_bar_at_zeta, PP_hat_at_zeta, QQ_bar_at_zeta, QQ_hat_at_zeta, &
       ZZ_at_zeta
  double complex, dimension(1:JJ-1,1:KK) :: PP_bar_at_u, PP_hat_at_u
  double complex, dimension(1:JJ,1:KK-1) :: QQ_bar_at_v, QQ_hat_at_v
  double complex PP_at_zeta(1:JJ), QQ_at_zeta(1:KK), &
       PP_at_u(1:JJ-1), QQ_at_v(1:KK-1) 
  double complex i, beta, delta, a1, b1, a2, b2, mu, nu, psi(1:NN), &
       AA_Bar(1:NN), kappa, gamma_a, lambda(1:NN), gamma(1:NN), &
       AA_Hat(1:NN)
  double precision c, pi

contains
    
! *****************************************************************

  subroutine setup_analytic

! *****************************************************************

    implicit none

    integer j, k, n, pp
    double precision x(1:NN), y(1:NN), x1
    double complex sum, LHS, mm(1:NN,1:NN+1), &
         XX(1:NN), ahold(1:NN,1:NN+1), a(1:NN,1:NN+1)

! setup some constants

    i = (0d0,1d0)

    pi = 4d0*atan(1d0)

    alpha = angle*pi/180d0

    c = sqrt(g*h0)

    sigma = 2d0*pi/period

    beta = sigma*i+Cl/h0**2

    delta = sqrt(beta*sigma*i/c**2)

    a1 = tau_0*cos(alpha)/(2d0*rho*c**2)  
    a2 = tau_0*sin(alpha)/(2d0*rho*c**2)
    b1 = 1d0/(1d0+exp(delta*LL))
    b2 = 1d0/(1d0+exp(delta*BB))

    mu = real(delta)
    nu = aimag(delta)

    kappa = sqrt(-4d0*sigma**2/c**2*(1d0-Cl*i/(2d0*h0**2*sigma)))

    gamma_a = sqrt(kappa**2-delta**2)

    do n=1,NN
       gamma(n) = dble(n)*pi*i/LL
       lambda(n) = sqrt(gamma(n)**2-kappa**2)
    end do

    do n=1,NN
       psi(n) = dble(n)*pi*i/LL
    end do

! specify values of x where Collacation technique is applied
    do j=1,NN
       x(j) = dble(j-1)*LL/dble(NN-1)
    end do

! setup equations to be solved for finding A_bar
    do j=1,NN
       do k=1,NN
          a(j,k) = coeff_A_Bar(x(j),k)
       end do
       a(j,NN+1) = RHS_Bar(x(j))
    end do

! solve these linear equations

    ahold = a

    do k=1,NN-1
       do j=k+1,NN
          mm(j,k) = a(j,k)/a(k,k)
          do pp=k+1,NN+1
             a(j,pp) = a(j,pp)-mm(j,k)*a(k,pp)
          end do
       end do
    end do

    do j=1,NN
       XX(j) = (0.0d0,0.0d0)
    end do

    XX(NN) = a(NN,NN+1)/a(NN,NN)

    do j=NN-1,1,-1
       sum = (0.0d0,0.0d0)
       do k=j+1,NN
          sum = sum + a(j,k)*XX(k)
       end do
       XX(j) = a(j,NN+1) - sum
       XX(j) = XX(j)/a(j,j)
    end do

    do j=1,NN
       sum = 0d0
       do k=1,NN
          sum = sum + ahold(j,k)*XX(k)
       end do
    end do
    
    AA_Bar(1:NN) = XX(1:NN)

!!$! Here we see if BC holds at other positions too (let's hope it does)
!!$      open(unit=1,file="lhs_bar.dat")
!!$      open(unit=2,file="rhs_bar.dat")
!!$      open(unit=3,file="diff_bar.dat")
!!$    do j=0,100
!!$       x1 = dble(j)*LL/100d0
!!$       LHS = 0d0
!!$       do n=1,NN
!!$          LHS = LHS + coeff_A_bar(x1,n)*AA_bar(n)
!!$       end do
!!$       write(1,*)real(LHS+conjg(LHS))
!!$       write(2,*)real(RHS_bar(x1)+conjg(RHS_bar(x1)))
!!$       write(3,*)real(RHS_bar(x1)+conjg(RHS_bar(x1)))-real(LHS+conjg(LHS))
!!$    end do
!!$    close(1)
!!$    close(2)
!!$    close(3)

! solve equations for A_hat

    do j=1,NN
       do k=1,NN
          a(j,k) = coeff_A_Hat(x(j),k)
       end do
       a(j,NN+1) = RHS_Hat(x(j))
    end do
      
    ahold = a

    do k=1,NN-1
       do j=k+1,NN
          mm(j,k) = a(j,k)/a(k,k)
          do pp=k+1,NN+1
             a(j,pp) = a(j,pp)-mm(j,k)*a(k,pp)
          end do
       end do
    end do
    
    do j=1,NN
       XX(j) = (0.0d0,0.0d0)
    end do
    
    XX(NN) = a(NN,NN+1)/a(NN,NN)

    do j=NN-1,1,-1
       sum = (0.0d0,0.0d0)
       do k=j+1,NN
          sum = sum + a(j,k)*XX(k)
       end do
       XX(j) = a(j,NN+1) - sum
       XX(j) = XX(j)/a(j,j)
    end do

! Here is where we check to see if matrix solver has worked

!!$    do j=1,NN
!!$       sum = 0d0
!!$       do k=1,NN
!!$          sum = sum + ahold(j,k)*XX(k)
!!$       end do
!!$       print*,sum, ahold(j,NN+1), sum-ahold(j,NN+1)
!!$       print*
!!$    end do

    AA_Hat(1:NN) = XX(1:NN)

!!$! Here we see if BC holds at other positions too (let's hope it does)
!!$      open(unit=1,file="lhs_hat.dat")
!!$      open(unit=2,file="rhs_hat.dat")
!!$      open(unit=3,file="diff_hat.dat")
!!$    do j=0,100
!!$       x1 = dble(j)*LL/100d0
!!$       LHS = 0d0
!!$       do n=1,NN
!!$          LHS = LHS + coeff_A_Hat(x1,n)*AA_Hat(n)
!!$       end do
!!$       write(1,*)real(LHS+conjg(LHS))
!!$       write(2,*)real(RHS_Hat(x1)+conjg(RHS_Hat(x1)))
!!$       write(3,*)real(RHS_Hat(x1)+conjg(RHS_Hat(x1)))-real(LHS+conjg(LHS))
!!$    end do
!!$    close(1)
!!$    close(2)
!!$    close(3)

! compute some term which will be used over and over again in the 
! analytic solution
    do j=1,JJ
       do k=1,KK
          ZZ_at_zeta(j,k) = ZZ(x_at_zeta(j),y_at_zeta(k))
          ZZ_bar_at_zeta(j,k) = ZZ_bar(x_at_zeta(j),y_at_zeta(k))
          ZZ_hat_at_zeta(j,k) = ZZ_hat(x_at_zeta(j),y_at_zeta(k))
          PP_at_zeta(j) = PP1(x_at_zeta(j))
          PP_bar_at_zeta(j,k) = PP_bar(x_at_zeta(j),y_at_zeta(k))
          PP_hat_at_zeta(j,k) = PP_hat(x_at_zeta(j),y_at_zeta(k))
          QQ_at_zeta(k) = QQ(y_at_zeta(k))
          QQ_bar_at_zeta(j,k) = QQ_bar(x_at_zeta(j),y_at_zeta(k))
          QQ_hat_at_zeta(j,k) = QQ_hat(x_at_zeta(j),y_at_zeta(k))
       end do
    end do
    do j=1,JJ-1
       do k=1,KK
          PP_at_u(j) = PP1(x_at_u(j))
          PP_bar_at_u(j,k) = PP_bar(x_at_u(j),y_at_u(k))
          PP_hat_at_u(j,k) = PP_hat(x_at_u(j),y_at_u(k))
       end do
    end do
    do j=1,JJ
       do k=1,KK-1
          QQ_at_v(k) = QQ(y_at_v(k))
          QQ_bar_at_v(j,k) = QQ_bar(x_at_v(j),y_at_v(k))
          QQ_hat_at_v(j,k) = QQ_hat(x_at_v(j),y_at_v(k))
       end do
    end do

    end subroutine setup_analytic

! *********************************************************
! 2nd-order analytic solution for zeta at zeta points (centre 
! of grid box)

  double precision function analytic_zeta_at_zeta(j,k,t)

! *********************************************************

    implicit none

    integer j,k
    double precision t
    double complex Z

    Z = ZZ_bar_at_zeta(j,k) &
         + ZZ_at_zeta(j,k)*exp(sigma*i*t) &
         + ZZ_hat_at_zeta(j,k)*exp(2d0*sigma*i*t)

    analytic_zeta_at_zeta = Z + conjg(Z)

    return

  end function analytic_zeta_at_zeta

! *********************************************************
! 2nd-order analytic solution for u at zeta points

  double precision function analytic_u_at_zeta(j,k,t)

! *********************************************************

    implicit none

    integer j,k
    double precision t
    double complex P

    P = PP_bar_at_zeta(j,k) &
         + PP_at_zeta(j)*exp(sigma*i*t) &
         + PP_hat_at_zeta(j,k)*exp(2d0*sigma*i*t)

    analytic_u_at_zeta = P + conjg(P)

    return

  end function analytic_u_at_zeta

! *********************************************************
! 2nd-order analytic solution for v at zeta points

  double precision function analytic_v_at_zeta(j,k,t)

! *********************************************************

    implicit none

    integer j,k
    double precision t
    double complex Q

    Q = QQ_bar_at_zeta(j,k) &
         + QQ_at_zeta(k)*exp(sigma*i*t) &
         + QQ_hat_at_zeta(j,k)*exp(2d0*sigma*i*t)

    analytic_v_at_zeta = Q + conjg(Q)

    return

  end function analytic_v_at_zeta

! *********************************************************
! 1st-order analytic solution for zeta at zeta points

  double precision function analytic_zeta1_at_zeta(j,k,t)

! *********************************************************

    implicit none

    integer j,k
    double precision t
    double complex Z

    Z = ZZ_at_zeta(j,k)*exp(sigma*i*t)

    analytic_zeta1_at_zeta = Z + conjg(Z)

    return

  end function analytic_zeta1_at_zeta

! *********************************************************
! 1st-order analytic solution for u at zeta points

  double precision function analytic_u1_at_zeta(j,t)

! *********************************************************

    implicit none

    integer j
    double precision t
    double complex P

    P = PP_at_zeta(j)*exp(sigma*i*t)

    analytic_u1_at_zeta = P + conjg(P)

    return

  end function analytic_u1_at_zeta

! *********************************************************
! 1st-order analytic solution for v at zeta points

  double precision function analytic_v1_at_zeta(k,t)

! *********************************************************

    implicit none

    integer k
    double precision t
    double complex Q

    Q = QQ_at_zeta(k)*exp(sigma*i*t)

    analytic_v1_at_zeta = Q + conjg(Q)

    return

  end function analytic_v1_at_zeta

! *********************************************************
! 2nd-order analytic solution for u at u points (midpoint 
! of rhs of grid box)

  double precision function analytic_u_at_u(j,k,t)

! *********************************************************

    implicit none

    integer j,k
    double precision t
    double complex P

    P = PP_bar_at_u(j,k) &
         + PP_at_u(j)*exp(sigma*i*t) &
         + PP_hat_at_u(j,k)*exp(2d0*sigma*i*t)

    analytic_u_at_u = P + conjg(P)

    return

  end function analytic_u_at_u

! *********************************************************
! 2nd-order analytic solution for v at v points (midpoint 
! of top of grid box)

  double precision function analytic_v_at_v(j,k,t)

! *********************************************************

    implicit none

    integer j,k
    double precision t
    double complex Q

    Q = QQ_bar_at_v(j,k) &
         + QQ_at_v(k)*exp(sigma*i*t) &
         + QQ_hat_at_v(j,k)*exp(2d0*sigma*i*t)

    analytic_v_at_v = Q + conjg(Q)

    return

  end function analytic_v_at_v

! *********************************************************
! 1st-order analytic solution for u at u points (midpoint 
! of rhs of grid box)

  double precision function analytic_u1_at_u(j,t)

! *********************************************************

    implicit none

    integer j
    double precision t
    double complex P

    P = PP_at_u(j)*exp(sigma*i*t)

    analytic_u1_at_u = P + conjg(P)

    return

  end function analytic_u1_at_u

! *********************************************************
! 1st-order analytic solution for v at v points (midpoint 
! of top of grid box)

  double precision function analytic_v1_at_v(k,t)

! *********************************************************

    implicit none

    integer k
    double precision t
    double complex Q

    Q = QQ_at_v(k)*exp(sigma*i*t)

    analytic_v1_at_v = Q + conjg(Q)

    return

  end function analytic_v1_at_v

! *********************************************************
! 2nd-order analytic solution for zeta at any location and time

  double precision function analytic_zeta(x,y,t)

! *********************************************************

    implicit none

    double precision x,y,t

    analytic_zeta = analytic_zeta1(x,y,t) + analytic_zeta2(x,y,t)

    return

  end function analytic_zeta

! *********************************************************
! 2nd-order analytic solution for u at any location and time

  double precision function analytic_u(x,y,t)

! *********************************************************

    implicit none

    double precision x,y,t

    analytic_u = analytic_u1(x,t) + analytic_u2(x,y,t)

    return

  end function analytic_u

! *********************************************************
! 2nd-order analytic solution for v at any location and time

  double precision function analytic_v(x,y,t)

! *********************************************************

    implicit none

    double precision x,y,t

    analytic_v = analytic_v1(y,t) + analytic_v2(x,y,t)

    return

  end function analytic_v

! *****************************************************************
! 1st-order analytic solution for zeta at any location and time

    double precision function analytic_zeta1(x,y,t)

! *****************************************************************

      double precision x, y, t

      analytic_zeta1 = ZZ(x,y)*exp(sigma*i*t) &
           + conjg(ZZ(x,y)*exp(sigma*i*t))

      return

    end function analytic_zeta1

! *****************************************************************
! 1st-order analytic solution for u at any location and time

    double precision function analytic_u1(x,t)

! *****************************************************************

      double precision x, t

      analytic_u1 = PP1(x)*exp(sigma*i*t) &
           + conjg(PP1(x)*exp(sigma*i*t))

      return

    end function analytic_u1

! *****************************************************************
! 1st-order analytic solution for v at any location and time

    double precision function analytic_v1(y,t)

! *****************************************************************

      double precision y, t

      analytic_v1 = QQ(y)*exp(sigma*i*t) &
           + conjg(QQ(y)*exp(sigma*i*t))

      return

    end function analytic_v1

! *****************************************************************
! 2nd-order component of analytic zeta 

    double precision function analytic_zeta2(x,y,t)

! *****************************************************************

      double precision x, y, t

      analytic_zeta2 = ZZ_Bar(x,y)+ZZ_hat(x,y)*exp(2d0*sigma*i*t) &
           + conjg(ZZ_Bar(x,y)+ZZ_hat(x,y)*exp(2d0*sigma*i*t))

      return

    end function analytic_zeta2

! *****************************************************************
! 2nd-order component of analytic u

    double precision function analytic_u2(x,y,t)

! *****************************************************************

      double precision x, y, t

      analytic_u2 = PP_Bar(x,y)+PP_hat(x,y)*exp(2d0*sigma*i*t) &
           + conjg(PP_Bar(x,y)+PP_hat(x,y)*exp(2d0*sigma*i*t))

      return

    end function analytic_u2

! *****************************************************************
! 2nd-order component of analytic v 

    double precision function analytic_v2(x,y,t)

! *****************************************************************

      double precision x, y, t

      analytic_v2 = QQ_bar(x,y)+QQ_hat(x,y)*exp(2d0*sigma*i*t) &
           + conjg(QQ_bar(x,y)+QQ_hat(x,y)*exp(2d0*sigma*i*t))

      return

    end function analytic_v2

! *****************************************************************
! coefficient function for 1st order zeta

    double complex function ZZ(x,y)

! *****************************************************************

      double precision x, y

      ZZ = 1d0/delta &
           *(a1*b1*(exp(delta*x)-exp(-delta*(x-LL))) &
           +a2*b2*(exp(delta*y)-exp(-delta*(y-BB))))

      return

    end function ZZ
    
! *****************************************************************
! coefficient function for 1st order U

    double complex function PP1(x)

! *****************************************************************

      double precision x

      PP1 = a1*c**2/beta &
           *(1d0-b1*(exp(delta*x)+exp(-delta*(x-LL))))

      return

    end function PP1

! *****************************************************************
! coefficient function for 1st order V

    double complex function QQ(y)

! *****************************************************************

      double precision y

      QQ = a2*c**2/beta &
           *(1d0-b2*(exp(delta*y)+exp(-delta*(y-BB))))

      return

    end function QQ
    
! *****************************************************************
! function for computing zeta_bar

    double complex function ZZ_bar(x,y)

! *****************************************************************

      double precision x, y

      ZZ_Bar = 1d0/(h0*delta*conjg(beta)) &
           *(-a1**2*abs(b1)**2*(conjg(beta)+2d0*beta)/(2d0*mu)&
           *(exp(2d0*mu*x)+exp(-2d0*mu*(x-LL))-1d0/(mu*LL) &
           *(exp(2d0*mu*LL)-1d0)) &
           +i/(2d0*nu)*a1**2*abs(b1)**2*(conjg(beta)+2d0*beta) &
           *(exp(2d0*i*nu*x+(mu-i*nu)*LL)+exp(-2d0*i*nu*x+(mu+i*nu)*LL)-i/(nu*LL) &
           *(exp((mu-i*nu)*LL)-exp((mu+i*nu)*LL))) &
           +2d0/delta*a1**2*b1*beta*(exp((mu+i*nu)*x)+exp(-(mu+i*nu)*(x-LL))-2d0/((mu+i*nu)*LL) &
           *(exp((mu+i*nu)*LL)-1d0)) &
           -1d0/(2d0*mu)*a2**2*abs(b2)**2*(conjg(beta)+2d0*beta) &
           *(exp(2d0*mu*y)+exp(-2d0*mu*(y-BB))-1d0/(mu*BB)*(exp(2d0*mu*BB)-1d0)) &
           +i/(2*nu)*a2**2*abs(b2)**2*(conjg(beta)+2d0*beta) &
           *(exp(2d0*i*nu*y+(mu-i*nu)*BB)+exp(-2d0*i*nu*y+(mu+i*nu)*BB)-i/(nu*BB) &
           *(exp((mu-i*nu)*BB)-exp((mu+i*nu)*BB))) &
           +2d0/delta*a2**2*b2*beta*(exp((mu+i*nu)*y)+exp(-(mu+i*nu)*(y-BB))-2d0/((mu+i*nu)*BB) &
           *(exp((mu+i*nu)*BB)-1d0)))
           
           ZZ_Bar = ZZ_Bar &
           -a1*a2*b2*i/(h0*delta**2*(1d0+exp(-(mu+i*nu)*i*LL))) &
           *(exp(-(mu+i*nu)*i*x)-exp((mu+i*nu)*i*(x-LL))) &
           *(exp((mu+i*nu)*y)-exp(-(mu+i*nu)*(y-BB)))
      
      do n=1,NN
         ZZ_Bar = ZZ_Bar + &
              AA_Bar(n)*(exp(psi(n)*x)+exp(-psi(n)*x)) &
              *(exp(-psi(n)*i*y)-exp(psi(n)*i*(y-BB)))
      end do

      return

    end function ZZ_bar
    
! *****************************************************************
! function for computing U_bar

    double complex function PP_bar(x,y)

! *****************************************************************

      double precision x, y

      PP_Bar = a1*a2*b2/(h0*delta*conjg(beta)) &
           *((2d0*beta-sigma*i)*(exp((mu+i*nu)*y)-exp(-(mu+i*nu)*(y-BB))) &
             -conjg(b1)*(2d0*beta+conjg(beta)-sigma*i) &
             *(exp((mu-i*nu)*x)+exp(-(mu-i*nu)*(x-LL))) &
             *(exp((mu+i*nu)*y)-exp(-(mu+i*nu)*(y-BB))) &
             +conjg(beta)/(1d0+exp(-(mu+i*nu)*i*LL)) &
             *(exp(-(mu+i*nu)*i*x)+exp((mu+i*nu)*i*(x-LL))) &
             *(exp((mu+i*nu)*y)-exp(-(mu+i*nu)*(y-BB))))
      
      do n=1,NN
         PP_Bar = PP_Bar &
              - AA_Bar(n)*psi(n)*(exp(psi(n)*x)-exp(-psi(n)*x)) &
              *(exp(-psi(n)*i*y)-exp(psi(n)*i*(y-BB)))
      end do

      PP_Bar = PP_Bar*h0**2*c**2/Cl
      
      return
      
    end function PP_bar

! *****************************************************************
! function for computing V_bar

    double complex function QQ_bar(x,y)

! *****************************************************************

      double precision x, y

      QQ_Bar = a2*a1/(h0*delta*conjg(beta)) &
           *(b1*(2d0*beta-sigma*i)*(exp((mu+i*nu)*x)-exp(-(mu+i*nu)*(x-LL))) &
                 -b1*conjg(b2)*(conjg(beta)+2d0*beta-sigma*i) &
                 *(exp((mu-i*nu)*y)+exp(-(mu-i*nu)*(y-BB))) &
                 *(exp((mu+i*nu)*x)-exp(-(mu+i*nu)*(x-LL))) &
                 +conjg(beta)*b2*i/(1d0+exp(-(mu+i*nu)*i*LL)) &
                 *(exp(-(mu+i*nu)*i*x)-exp((mu+i*nu)*i*(x-LL))) &
                 *(exp((mu+i*nu)*y)+exp(-(mu+i*nu)*(y-BB))))
      
      do n=1,NN
         QQ_Bar = QQ_Bar &
              + psi(n)*i*AA_bar(n)*(exp(psi(n)*x)+exp(-psi(n)*x)) &
              *(exp(-psi(n)*i*y)+exp(psi(n)*i*(y-BB)))

      end do

      QQ_Bar = QQ_Bar*h0**2*c**2/Cl
      
      return
      
    end function QQ_bar

! *****************************************************************
! function for computing zeta_hat

    double complex function ZZ_hat(x,y)

! *****************************************************************

      double precision x, y

      ZZ_hat = -3d0*a1**2*b1**2/(beta*i+sigma)*(exp(2d0*delta*x)+exp(-2d0*delta*(x-LL))) &
           -3d0*a2**2*b2**2/(beta*i+sigma)*(exp(2d0*delta*y)+exp(-2d0*delta*(y-BB))) &
           +2d0*a1**2*b1/(-beta*i+2d0*sigma)*(exp(delta*x)+exp(-delta*(x-LL))) &
           +2d0*a2**2*b2/(-beta*i+2d0*sigma)*(exp(delta*y)+exp(-delta*(y-BB))) &
           -a1*a2*b1*b2*(3d0*beta-sigma*i)/(beta*sigma) &
           *(exp(delta*x)-exp(-delta*(x-LL)))*(exp(delta*y)-exp(-delta*(y-BB)))

      ZZ_hat = ZZ_hat*c**2/(h0*sigma)

      ZZ_hat = ZZ_hat &
           +a1**2*b1*(beta*i-sigma)/(h0*delta*kappa &
           *(1d0-exp(kappa*LL)))*(3d0*b1/(beta*i+sigma) &
           *(1d0-exp(2d0*delta*LL))-4d0/(2d0*sigma-beta*i) &
           *(1d0-exp(delta*LL)))*(exp(kappa*x)+exp(-kappa*(x-LL))) &
           +3d0*beta*i*a1*a2*b2/(h0*delta*sigma*gamma_a &
           *(1d0+exp(gamma_a*LL)))*(exp(gamma_a*x)-exp(-gamma_a*(x-LL))) &
           *(exp(delta*y)-exp(-delta*(y-BB))) &
           +a2**2*b2*(beta*i-sigma)/(h0*delta*kappa*(1d0-exp(kappa*BB))) &
           *(3d0*b2/(beta*i+sigma)*(1d0-exp(2d0*delta*BB))-4d0/(2d0*sigma-beta*i) &
           *(1d0-exp(delta*BB)))*(exp(kappa*y)+exp(-kappa*(y-BB)))

      do n=1,NN
         ZZ_hat = ZZ_hat &
              +AA_hat(n)*(exp(gamma(n)*x)+exp(-gamma(n)*x)) &
              *(exp(-lambda(n)*i*y)-exp(lambda(n)*i*(y-BB)))
      end do

      return

    end function ZZ_hat

! *****************************************************************
! function for computing U_hat

    double complex function PP_hat(x,y)

! *****************************************************************

      double precision x, y

      PP_hat = 3d0*a1*b1**2*(beta*i-sigma)/(beta*i+sigma) &
           *(exp(2d0*delta*x)-exp(-2d0*delta*(x-LL))) &
           +4d0*a1*b1*(beta*i-sigma)/(beta*i-2d0*sigma) &
           *(exp(delta*x)-exp(-delta*(x-LL))) &
           +a2*b2*(2d0*beta-sigma*i)/beta*(exp(delta*y)-exp(-delta*(y-BB))) &
           +a2*b1*b2*(3d0*beta-sigma*i)*(beta*i-sigma)/(beta*sigma) &
           *(exp(delta*x)+exp(-delta*(x-LL)))*(exp(delta*y)-exp(-delta*(y-BB))) &
           -a1*b1*(beta*i-sigma)/(1d0-exp(kappa*LL))*(3d0*b1/(beta*i+sigma) &
           *(1d0-exp(2d0*delta*LL))-4d0/(2d0*sigma-beta*i)*(1d0-exp(delta*LL))) &
           *(exp(kappa*x)-exp(-kappa*(x-LL))) &
           -3d0*beta*i*a2*b2/(sigma*(1d0+exp(gamma_a*LL))) &
           *(exp(gamma_a*x)+exp(-gamma_a*(x-LL)))*(exp(delta*y)-exp(-delta*(y-BB)))

      PP_hat = PP_hat*a1/(h0*delta)

      do n=1,NN
         PP_hat = PP_hat &
              -AA_hat(n)*gamma(n)*(exp(gamma(n)*x)-exp(-gamma(n)*x)) &
              *(exp(-lambda(n)*i*y)-exp(lambda(n)*i*(y-BB)))
      end do

      PP_hat = PP_hat*c**2/(beta+sigma*i)

      return

    end function PP_hat

! *****************************************************************
! function for computing V_hat

    double complex function QQ_hat(x,y)

! *****************************************************************

      double precision x, y

      QQ_hat = 3d0*a2*b2**2*(beta*i-sigma)/(beta*i+sigma)*(exp(2d0*delta*y)-exp(-2d0*delta*(y-BB))) &
           +4d0*a2*b2*(beta*i-sigma)/(beta*i-2d0*sigma)*(exp(delta*y)-exp(-delta*(y-BB))) &
           +a1*b1*(2d0*beta-sigma*i)/beta*(exp(delta*x)-exp(-delta*(x-LL))) &
           +a1*b1*b2*(3d0*beta-sigma*i)*(beta*i-sigma)/(beta*sigma) &
           *(exp(delta*x)-exp(-delta*(x-LL)))*(exp(delta*y)+exp(-delta*(y-BB))) &
           -a2*b2*(beta*i-sigma)/(1d0-exp(kappa*BB))*(3d0*b2/(beta*i+sigma) &
           *(1d0-exp(2d0*delta*BB))-4d0/(2d0*sigma-beta*i)*(1d0-exp(delta*BB))) &
           *(exp(kappa*y)-exp(-kappa*(y-BB))) &
           -3d0*beta*i*a1*b2*delta/(sigma*gamma_a*(1d0+exp(gamma_a*LL))) &
           *(exp(gamma_a*x)-exp(-gamma_a*(x-LL)))*(exp(delta*y)+exp(-delta*(y-BB)))

      QQ_hat = QQ_hat*a2/(h0*delta)

      do n=1,NN
         QQ_hat = QQ_hat &
              +i*AA_hat(n)*lambda(n)*(exp(gamma(n)*x)+exp(-gamma(n)*x)) &
              *(exp(-lambda(n)*i*y)+exp(lambda(n)*i*(y-BB)))

      end do

      QQ_hat = QQ_hat*c**2/(beta+sigma*i)

      return

    end function QQ_hat

! *****************************************************************
! coefficients for calculating A_bar

    double complex function coeff_A_Bar(x,n)

! *****************************************************************

      double precision x
      integer n

      coeff_A_Bar = psi(n)*i*(exp(psi(n)*x)+exp(-psi(n)*x)) &
           *(1d0+exp(-psi(n)*i*BB))

      return

    end function coeff_A_Bar

! *****************************************************************
! rhs of equations for computing A_bar

    double complex function RHS_Bar(x)

! *****************************************************************

      double precision x

      RHS_Bar = a1*a2/(h0*delta) &
           *(b1*(exp((mu+i*nu)*x)-exp(-(mu+i*nu)*(x-LL))) &
             -i/(1d0+exp(-(mu+i*nu)*i*LL)) &
                 *(exp(-(mu+i*nu)*i*x)-exp((mu+i*nu)*i*(x-LL))))

      return

    end function RHS_Bar

! *****************************************************************
! coefficients for calculating A_hat

    double complex function coeff_A_Hat(x,n)

! *****************************************************************

      double precision x
      integer n

      coeff_A_Hat = lambda(n)*(exp(gamma(n)*x)+exp(-gamma(n)*x))*(1d0+exp(-lambda(n)*i*BB))

      return

    end function coeff_A_Hat

! *****************************************************************
! rhs of equations for computing A_hat

    double complex function RHS_Hat(x)

! *****************************************************************

      double precision x

      RHS_Hat = -3d0*beta*a1*a2/(h0*sigma)*(b1/delta &
           *(exp(delta*x)-exp(-delta*(x-LL)))-1d0/(gamma_a*(1d0+exp(gamma_a*LL))) &
           *(exp(gamma_a*x)-exp(-gamma_a*(x-LL))))

      return

    end function RHS_Hat

! *****************************************************************

  end module analytic
  
