! *****************************************************************
!
! This file contains a Fortran 90 module for outputting analytic and 
! numerical data to ascii and NetCDF files
!
! Copyright (C) 2009 David McInerney, University of Chicago
!
! *****************************************************************
!
!    This program is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with this program (see file "COPYING").  If not, see
!    <http://www.gnu.org/licenses/>.
!
! *****************************************************************

module output_module

  use grid, only : JJ, KK, NN, &
       x_at_zeta, y_at_zeta, x_at_u, y_at_u, x_at_v, y_at_v, t

  implicit none

#ifdef HAVE_NETCDF
  include 'netcdf.inc'
#endif

  character (len=50) :: dirname
  integer j1,k1,j2,k2,j3,k3,j4,k4,j5,k5,j6,k6

  double precision :: x_ncgrid(0:2*JJ), y_ncgrid(0:2*KK)
  double precision, dimension(0:2*JJ,0:2*KK) :: &
       analytic_zeta_nc, analytic_u_nc, analytic_v_nc, &
       analytic_zeta1_nc, analytic_u1_nc, analytic_v1_nc, &
       analytic_zeta2_nc, analytic_u2_nc, analytic_v2_nc       
  integer ncid, x_at_zeta_dim_id, y_at_zeta_dim_id, x_at_u_dim_id, y_at_u_dim_id, &
       x_at_v_dim_id, y_at_v_dim_id, &
       analytic_zeta_id, analytic_u_id, analytic_v_id, &
       time_dim_id, &
       numerical_u_id, numerical_v_id, numerical_zeta_id, &
       x_ncgrid_dim_id, y_ncgrid_dim_id, x_ncgrid_id, y_ncgrid_id, &
       x_at_zeta_id, y_at_zeta_id, x_at_u_id, y_at_u_id, x_at_v_id, y_at_v_id, &
       time_id, &
       analytic_zeta1_id, analytic_u1_id, analytic_v1_id, &
       analytic_zeta2_id, analytic_u2_id, analytic_v2_id      
  integer :: output_per_period = 100

  integer time_out_counter

contains

! ***************************************************************

  subroutine open_files

! ***************************************************************

    use parameters, only : LL, BB, no_warmups

    implicit none

    integer, dimension(3) :: analytic_dims, zeta_dims, u_dims, v_dims
    integer rcode
    integer j,k

    time_out_counter = 0
    
    do j = 0,2*JJ
       x_ncgrid(j) = dble(j)*LL/(2d0*dble(JJ))
    end do
    do k = 0,2*KK
       y_ncgrid(k) = dble(k)*BB/(2d0*dble(KK))
    end do

#ifdef HAVE_NETCDF

    ncid = nccre(trim(dirname)//"/output.nc", NCCLOB, rcode)

    ! create dimensions
    x_at_zeta_dim_id = ncddef(ncid,'x_at_zeta',JJ,rcode)
    y_at_zeta_dim_id = ncddef(ncid,'y_at_zeta',KK,rcode)
    x_at_u_dim_id = ncddef(ncid,'x_at_u',JJ+1,rcode)
    y_at_u_dim_id = ncddef(ncid,'y_at_u',KK,rcode)
    x_at_v_dim_id = ncddef(ncid,'x_at_v',JJ,rcode)
    y_at_v_dim_id = ncddef(ncid,'y_at_v',KK+1,rcode)
    x_ncgrid_dim_id = ncddef(ncid,'x',2*JJ+1,rcode)
    y_ncgrid_dim_id = ncddef(ncid,'y',2*KK+1,rcode)
    time_dim_id = ncddef(ncid,'time',ncunlim,rcode)

    ! create variables that correspond to dimensions
    x_at_zeta_id = ncvdef(ncid,'x_at_zeta',NCDOUBLE,1,x_at_zeta_dim_id,rcode)
    y_at_zeta_id = ncvdef(ncid,'y_at_zeta',NCDOUBLE,1,y_at_zeta_dim_id,rcode)
    x_at_u_id = ncvdef(ncid,'x_at_u',NCDOUBLE,1,x_at_u_dim_id,rcode)
    y_at_u_id = ncvdef(ncid,'y_at_u',NCDOUBLE,1,y_at_u_dim_id,rcode)
    x_at_v_id = ncvdef(ncid,'x_at_v',NCDOUBLE,1,x_at_v_dim_id,rcode)
    y_at_v_id = ncvdef(ncid,'y_at_v',NCDOUBLE,1,y_at_v_dim_id,rcode)
    x_ncgrid_id = ncvdef(ncid,'x',NCDOUBLE,1,x_ncgrid_dim_id,rcode)
    y_ncgrid_id = ncvdef(ncid,'y',NCDOUBLE,1,y_ncgrid_dim_id,rcode)
    time_id = ncvdef(ncid,'time',NCDOUBLE,1,time_dim_id,rcode)

    ! create analytic variables
    analytic_dims(1) = x_ncgrid_dim_id
    analytic_dims(2) = y_ncgrid_dim_id
    analytic_dims(3) = time_dim_id

    analytic_zeta_id = ncvdef(ncid,'Analytic_zeta',NCDOUBLE,3,analytic_dims,rcode)
    analytic_zeta1_id = ncvdef(ncid,'Analytic_zeta1',NCDOUBLE,3,analytic_dims,rcode)
    analytic_zeta2_id = ncvdef(ncid,'Analytic_zeta2',NCDOUBLE,3,analytic_dims,rcode)
    
    analytic_u_id = ncvdef(ncid,'Analytic_U',NCDOUBLE,3,analytic_dims,rcode)
    analytic_u1_id = ncvdef(ncid,'Analytic_U1',NCDOUBLE,3,analytic_dims,rcode)
    analytic_u2_id = ncvdef(ncid,'Analytic_U2',NCDOUBLE,3,analytic_dims,rcode)

    analytic_v_id = ncvdef(ncid,'Analytic_V',NCDOUBLE,3,analytic_dims,rcode)
    analytic_v1_id = ncvdef(ncid,'Analytic_V1',NCDOUBLE,3,analytic_dims,rcode)
    analytic_v2_id = ncvdef(ncid,'Analytic_V2',NCDOUBLE,3,analytic_dims,rcode)

    ! create numerical variables
    zeta_dims(1) = x_at_zeta_dim_id
    zeta_dims(2) = y_at_zeta_dim_id
    zeta_dims(3) = time_dim_id
    numerical_zeta_id = ncvdef(ncid,'Numerical_zeta',NCDOUBLE,3,zeta_dims,rcode)

    u_dims(1) = x_at_u_dim_id
    u_dims(2) = y_at_u_dim_id
    u_dims(3) = time_dim_id
    numerical_u_id = ncvdef(ncid,'Numerical_U',NCDOUBLE,3,u_dims,rcode)

    v_dims(1) = x_at_v_dim_id
    v_dims(2) = y_at_v_dim_id
    v_dims(3) = time_dim_id
    numerical_v_id = ncvdef(ncid,'Numerical_V',NCDOUBLE,3,v_dims,rcode)

    call ncendf(ncid,rcode)

    call ncvpt(ncid,x_at_zeta_id, (/ 1 /), (/ JJ /), x_at_zeta, rcode)
    call ncvpt(ncid,y_at_zeta_id, (/ 1 /), (/ KK /), y_at_zeta, rcode)
    call ncvpt(ncid,x_at_u_id, (/ 1 /), (/ JJ+1 /), x_at_u, rcode)
    call ncvpt(ncid,y_at_u_id, (/ 1 /), (/ KK /), y_at_u, rcode)
    call ncvpt(ncid,x_at_v_id, (/ 1 /), (/ JJ /), x_at_v, rcode)
    call ncvpt(ncid,y_at_v_id, (/ 1 /), (/ KK+1 /), y_at_v, rcode)
    call ncvpt(ncid,x_ncgrid_id, (/ 1 /), (/ 2*JJ+1 /), x_ncgrid, rcode)
    call ncvpt(ncid,y_ncgrid_id, (/ 1 /), (/ 2*KK+1 /), y_ncgrid, rcode)

#endif

    open(unit=1001,file=trim(dirname)// "/" // "progress.txt")

  end subroutine open_files

! ***************************************************************

  subroutine write_zeta_u_v

! ***************************************************************

    use analytic, only : analytic_zeta_at_zeta, analytic_zeta1_at_zeta, &
         analytic_u_at_u, analytic_u1_at_u, &
         analytic_v_at_v, analytic_v1_at_v, &
         analytic_zeta, analytic_u, analytic_v, &
         analytic_zeta1, analytic_u1, analytic_v1, &
         analytic_zeta2, analytic_u2, analytic_v2

    use parameters, only : no_warmups, period
    use grid, only : t, zeta, u_new, v_new, JJ, KK, iteration

    implicit none

    integer j,k, rcode
    integer, dimension(3) :: start, size

#ifdef HAVE_NETCDF   

    time_out_counter = time_out_counter + 1

    start = (/ 1, 1, time_out_counter /)
    size = (/ JJ, KK, 1 /)
    call ncvpt(ncid,numerical_zeta_id, start, size, zeta, rcode)

    start = (/ 1, 1, time_out_counter /)
    size = (/ JJ+1, KK, 1 /)
    call ncvpt(ncid,numerical_u_id, start, size, u_new, rcode)

    start = (/ 1, 1, time_out_counter /)
    size = (/ JJ, KK+1, 1 /)
    call ncvpt(ncid,numerical_v_id, start, size, v_new, rcode)

    do j = 0,2*JJ
       do k = 0,2*KK
          analytic_zeta_nc(j,k) &
               = analytic_zeta(x_ncgrid(j),y_ncgrid(k),t)
          analytic_u_nc(j,k) &
               = analytic_u(x_ncgrid(j),y_ncgrid(k),t)
          analytic_v_nc(j,k) &
               = analytic_v(x_ncgrid(j),y_ncgrid(k),t)
          analytic_zeta1_nc(j,k) &
               = analytic_zeta1(x_ncgrid(j),y_ncgrid(k),t)
          analytic_u1_nc(j,k) &
               = analytic_u1(x_ncgrid(j),t)
          analytic_v1_nc(j,k) &
               = analytic_v1(y_ncgrid(k),t)
          analytic_zeta2_nc(j,k) &
               = analytic_zeta2(x_ncgrid(j),y_ncgrid(k),t)
          analytic_u2_nc(j,k) &
               = analytic_u2(x_ncgrid(j),y_ncgrid(k),t)
          analytic_v2_nc(j,k) &
               = analytic_v2(x_ncgrid(j),y_ncgrid(k),t)
       end do
    end do

    start = (/ 1, 1, time_out_counter /)
    size = (/ 2*JJ+1, 2*KK+1, 1 /)

    call ncvpt(ncid, analytic_zeta_id, start, size,analytic_zeta_nc, rcode)
    call ncvpt(ncid, analytic_u_id, start, size, analytic_u_nc, rcode)
    call ncvpt(ncid, analytic_v_id, start, size, analytic_v_nc, rcode)
    call ncvpt(ncid, analytic_zeta1_id, start, size, analytic_zeta1_nc, rcode)
    call ncvpt(ncid, analytic_u1_id, start, size, analytic_u1_nc, rcode)
    call ncvpt(ncid, analytic_v1_id, start, size, analytic_v1_nc, rcode)
    call ncvpt(ncid, analytic_zeta2_id, start, size, analytic_zeta2_nc, rcode)
    call ncvpt(ncid, analytic_u2_id, start, size, analytic_u2_nc, rcode)
    call ncvpt(ncid, analytic_v2_id, start, size, analytic_v2_nc, rcode)

    call ncvpt(ncid, time_id, (/ time_out_counter /), (/ 1 /), (t-dble(no_warmups)*period)/60d0/60d0, rcode)

#endif

  end subroutine write_zeta_u_v

! ***************************************************************

end module output_module
